"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Functions = undefined;

var _completions = require("../completions.js");

var _errors = require("../errors.js");

var _invariant = require("../invariant.js");

var _invariant2 = _interopRequireDefault(_invariant);

var _realm = require("../realm.js");

var _errors2 = require("../utils/errors.js");

var _index = require("../values/index.js");

var _generator = require("../utils/generator.js");

var _index2 = require("../methods/index.js");

var _modules = require("../utils/modules.js");

var _singletons = require("../singletons.js");

var _types = require("./types");

var _reconcilation = require("../react/reconcilation.js");

var _utils = require("../react/utils.js");

var _babelTypes = require("babel-types");

var t = _interopRequireWildcard(_babelTypes);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Functions {
  constructor(realm, moduleTracer) {
    this.realm = realm;
    this.moduleTracer = moduleTracer;
    this.writeEffects = new Map();
    this.functionExpressions = new Map();
    this._noOpFunction = undefined;
  }
  // maps back from FunctionValue to the expression string


  __optimizedFunctionEntryOfValue(value) {
    let realm = this.realm;
    if (value instanceof _index.ECMAScriptSourceFunctionValue) {
      // additional function logic
      return { value };
    } else if (value instanceof _index.ObjectValue) {
      // React component tree logic
      let config = (0, _index2.Get)(realm, value, "config");
      let rootComponent = (0, _index2.Get)(realm, value, "rootComponent");
      let validConfig = config instanceof _index.ObjectValue || config === realm.intrinsics.undefined;
      let validRootComponent = rootComponent instanceof _index.ECMAScriptSourceFunctionValue || rootComponent instanceof _index.AbstractValue && (0, _utils.valueIsKnownReactAbstraction)(this.realm, rootComponent);

      if (validConfig && validRootComponent) {
        return {
          value: rootComponent,
          config: (0, _utils.convertConfigObjectToReactComponentTreeConfig)(realm, config)
        };
      }
    }

    let location = value.expressionLocation ? `${value.expressionLocation.start.line}:${value.expressionLocation.start.column} ` + `${value.expressionLocation.end.line}:${value.expressionLocation.end.line}` : "location unknown";
    realm.handleError(new _errors.CompilerDiagnostic(`Optimized Function Value ${location} is an not a function or react element`, undefined, "PP0033", "FatalError"));
    throw new _errors.FatalError("Optimized Function Values must be functions or react elements");
  }

  __generateInitialAdditionalFunctions(globalKey) {
    let recordedAdditionalFunctions = [];
    let realm = this.realm;
    let globalRecordedAdditionalFunctionsMap = this.moduleTracer.modules.logger.tryQuery(() => (0, _index2.Get)(realm, realm.$GlobalObject, globalKey), realm.intrinsics.undefined);
    (0, _invariant2.default)(globalRecordedAdditionalFunctionsMap instanceof _index.ObjectValue);
    for (let funcId of globalRecordedAdditionalFunctionsMap.getOwnPropertyKeysArray()) {
      let property = globalRecordedAdditionalFunctionsMap.properties.get(funcId);
      if (property) {
        let value = property.descriptor && property.descriptor.value;
        (0, _invariant2.default)(value !== undefined);
        (0, _invariant2.default)(value instanceof _index.Value);
        let entry = this.__optimizedFunctionEntryOfValue(value);
        if (entry) recordedAdditionalFunctions.push(entry);
      }
    }
    return recordedAdditionalFunctions;
  }

  // NB: effects that are returned may be different than the effects passed in, so after this call, you may no longer
  // use the effects object you passed into this function.
  _createAdditionalEffects(effects, fatalOnAbrupt, name, environmentRecordIdAfterGlobalCode, parentAdditionalFunction = undefined) {
    let realm = this.realm;
    let [result, generator] = effects.data;
    if (result instanceof _completions.PossiblyNormalCompletion || result instanceof _completions.JoinedAbruptCompletions) {
      // joined generators have joined entries that will get visited recursively via result, so get rid of them here
      generator.purgeEntriesWithGeneratorDepencies();
      // The completion is not the end of function execution, but a fork point for separate threads of control.
      // The effects of all of these threads need to get joined up and rolled into the top level effects,
      // so that applying the effects before serializing the body will fully initialize all variables and objects.
      effects = realm.evaluateForEffects(() => {
        realm.applyEffects(effects, "_createAdditionalEffects/1", true);
        if (result instanceof _completions.PossiblyNormalCompletion) {
          result = _singletons.Join.joinPossiblyNormalCompletionWithAbruptCompletion(realm, result, new _completions.ReturnCompletion(result.value), (0, _realm.construct_empty_effects)(realm)).result;
        }
        (0, _invariant2.default)(result instanceof _completions.JoinedAbruptCompletions);
        let completionEffects = _singletons.Join.joinNestedEffects(realm, result);
        realm.applyEffects(completionEffects, "_createAdditionalEffects/2", false);
        return result;
      }, undefined, "_createAdditionalEffects");
    }
    let retValue = {
      parentAdditionalFunction,
      effects,
      transforms: [],
      generator: _generator.Generator.fromEffects(effects, this.realm, name, environmentRecordIdAfterGlobalCode),
      additionalRoots: new Set()
    };
    return retValue;
  }

  _generateWriteEffectsForReactComponentTree(componentType, effects, componentTreeState, evaluatedNode, environmentRecordIdAfterGlobalCode) {
    let additionalFunctionEffects = this._createAdditionalEffects(effects, false, "ReactAdditionalFunctionEffects", environmentRecordIdAfterGlobalCode);
    if (additionalFunctionEffects === null) {
      // TODO we don't support this yet, but will do very soon
      // to unblock work, we'll just return at this point right now
      evaluatedNode.status = "UNSUPPORTED_COMPLETION";
      return;
    }
    effects = additionalFunctionEffects.effects;
    let value = effects.result;

    if (value === this.realm.intrinsics.undefined) {
      // if we get undefined, then this component tree failed and a message was already logged
      // in the reconciler
      return;
    }
    if ((0, _utils.valueIsClassComponent)(this.realm, componentType)) {
      if (componentTreeState.status === "SIMPLE") {
        // if the root component was a class and is now simple, we can convert it from a class
        // component to a functional component
        (0, _utils.convertSimpleClassComponentToFunctionalComponent)(this.realm, componentType, additionalFunctionEffects);
        (0, _utils.normalizeFunctionalComponentParamaters)(componentType);
        this.writeEffects.set(componentType, additionalFunctionEffects);
      } else {
        let prototype = (0, _index2.Get)(this.realm, componentType, "prototype");
        (0, _invariant2.default)(prototype instanceof _index.ObjectValue);
        let renderMethod = (0, _index2.Get)(this.realm, prototype, "render");
        (0, _invariant2.default)(renderMethod instanceof _index.ECMAScriptSourceFunctionValue);
        this.writeEffects.set(renderMethod, additionalFunctionEffects);
      }
    } else {
      if (componentTreeState.status === "COMPLEX") {
        (0, _utils.convertFunctionalComponentToComplexClassComponent)(this.realm, componentType, componentTreeState.componentType, additionalFunctionEffects);
        let prototype = (0, _index2.Get)(this.realm, componentType, "prototype");
        (0, _invariant2.default)(prototype instanceof _index.ObjectValue);
        let renderMethod = (0, _index2.Get)(this.realm, prototype, "render");
        (0, _invariant2.default)(renderMethod instanceof _index.ECMAScriptSourceFunctionValue);
        this.writeEffects.set(renderMethod, additionalFunctionEffects);
      } else {
        (0, _utils.normalizeFunctionalComponentParamaters)(componentType);
        this.writeEffects.set(componentType, additionalFunctionEffects);
      }
    }
    // apply contextTypes for legacy context
    if (componentTreeState.contextTypes.size > 0) {
      let contextTypes = new _index.ObjectValue(this.realm, this.realm.intrinsics.ObjectPrototype);
      let noOpFunc = this._getNoOpFunction();
      for (let key of componentTreeState.contextTypes) {
        _singletons.Properties.Set(this.realm, contextTypes, key, noOpFunc, true);
      }
      _singletons.Properties.Set(this.realm, componentType, "contextTypes", contextTypes, true);
    }
  }

  _getNoOpFunction() {
    if (this._noOpFunction) {
      return this._noOpFunction;
    }
    let noOpFunc = new _index.ECMAScriptSourceFunctionValue(this.realm);
    let body = t.blockStatement([]);
    body.uniqueOrderedTag = this.realm.functionBodyUniqueTagSeed++;
    noOpFunc.$FormalParameters = [];
    noOpFunc.$ECMAScriptCode = body;
    this._noOpFunction = noOpFunc;
    return noOpFunc;
  }

  _forEachBindingOfEffects(effects, func) {
    let [result,, nestedBindingsToIgnore] = effects.data;
    for (let [binding] of nestedBindingsToIgnore) {
      func(binding);
    }
    if (result instanceof _completions.PossiblyNormalCompletion || result instanceof _completions.JoinedAbruptCompletions) {
      this._forEachBindingOfEffects(result.alternateEffects, func);
      this._forEachBindingOfEffects(result.consequentEffects, func);
    }
  }

  _hasWriteConflictsFromReactRenders(bindings, effects, nestedEffectsList, evaluatedNode) {
    let ignoreBindings = new Set();
    let failed = false;
    // TODO: should we also check realm.savedEffects?
    // ref: https://github.com/facebook/prepack/pull/1742

    for (let nestedEffects of nestedEffectsList) {
      this._forEachBindingOfEffects(nestedEffects, binding => {
        ignoreBindings.add(binding);
      });
    }

    this._forEachBindingOfEffects(effects, binding => {
      if (bindings.has(binding) && !ignoreBindings.has(binding)) {
        failed = true;
      }
      bindings.add(binding);
    });

    if (failed) {
      evaluatedNode.status = "WRITE-CONFLICTS";
    }
    return failed;
  }

  optimizeReactComponentTreeRoots(statistics, react, environmentRecordIdAfterGlobalCode) {
    let logger = this.moduleTracer.modules.logger;
    let bindings = new Set();
    let recordedReactRootValues = this.__generateInitialAdditionalFunctions("__reactComponentTrees");
    // Get write effects of the components
    if (this.realm.react.verbose) {
      logger.logInformation(`Evaluating ${recordedReactRootValues.length} React component tree roots...`);
    }
    for (let _ref of recordedReactRootValues) {
      let { value: componentRoot, config } = _ref;

      (0, _invariant2.default)(config);
      let reconciler = new _reconcilation.Reconciler(this.realm, this.moduleTracer.modules.logger, statistics, react, config);
      let componentType = (0, _utils.getComponentTypeFromRootValue)(this.realm, componentRoot);
      if (componentType === null) {
        continue;
      }
      let evaluatedRootNode = (0, _utils.createReactEvaluatedNode)("ROOT", (0, _utils.getComponentName)(this.realm, componentType));
      statistics.evaluatedRootNodes.push(evaluatedRootNode);
      if (reconciler.hasEvaluatedRootNode(componentType, evaluatedRootNode)) {
        continue;
      }
      let componentTreeEffects = reconciler.renderReactComponentTree(componentType, null, null, evaluatedRootNode);
      if (componentTreeEffects === null) {
        if (this.realm.react.verbose) {
          logger.logInformation(`  ✖ ${evaluatedRootNode.name} (root)`);
        }
        continue;
      }
      if (this._hasWriteConflictsFromReactRenders(bindings, componentTreeEffects, [], evaluatedRootNode)) {
        if (this.realm.react.verbose) {
          logger.logInformation(`  ✖ ${evaluatedRootNode.name} (root - write conflicts)`);
        }
        continue;
      }
      if (this.realm.react.verbose) {
        logger.logInformation(`  ✔ ${evaluatedRootNode.name} (root)`);
      }

      this._generateWriteEffectsForReactComponentTree(componentType, componentTreeEffects, reconciler.componentTreeState, evaluatedRootNode, environmentRecordIdAfterGlobalCode);

      let startingComponentTreeBranches = 0;
      do {
        startingComponentTreeBranches = reconciler.branchedComponentTrees.length;
        this._optimizeReactComponentTreeBranches(reconciler, bindings, environmentRecordIdAfterGlobalCode);
        this._optimizeReactNestedClosures(reconciler, bindings, environmentRecordIdAfterGlobalCode);
      } while (startingComponentTreeBranches !== reconciler.branchedComponentTrees.length);
    }
  }

  _optimizeReactNestedClosures(reconciler, bindings, environmentRecordIdAfterGlobalCode) {
    let logger = this.moduleTracer.modules.logger;

    if (this.realm.react.verbose && reconciler.nestedOptimizedClosures.length > 0) {
      logger.logInformation(`  Evaluating nested closures...`);
    }
    for (let _ref2 of reconciler.nestedOptimizedClosures) {
      let {
        func,
        evaluatedNode,
        nestedEffects,
        componentType,
        context,
        branchState
      } = _ref2;

      if (reconciler.hasEvaluatedNestedClosure(func)) {
        continue;
      }
      if (func instanceof _index.ECMAScriptSourceFunctionValue && reconciler.hasEvaluatedRootNode(func, evaluatedNode)) {
        continue;
      }
      let closureEffects = reconciler.renderNestedOptimizedClosure(func, nestedEffects, componentType, context, branchState, evaluatedNode);
      if (closureEffects === null) {
        if (this.realm.react.verbose) {
          logger.logInformation(`    ✖ function "${(0, _utils.getComponentName)(this.realm, func)}"`);
        }
        continue;
      }
      if (this._hasWriteConflictsFromReactRenders(bindings, closureEffects, nestedEffects, evaluatedNode)) {
        if (this.realm.react.verbose) {
          logger.logInformation(`    ✖ function "${(0, _utils.getComponentName)(this.realm, func)}" (write conflicts)`);
        }
        continue;
      }
      if (this.realm.react.verbose) {
        logger.logInformation(`    ✔ function "${(0, _utils.getComponentName)(this.realm, func)}"`);
      }
      let additionalFunctionEffects = this._createAdditionalEffects(closureEffects, true, "ReactNestedAdditionalFunctionEffects", environmentRecordIdAfterGlobalCode);
      (0, _invariant2.default)(additionalFunctionEffects);
      if (func instanceof _index.BoundFunctionValue) {
        (0, _invariant2.default)(func.$BoundTargetFunction instanceof _index.ECMAScriptSourceFunctionValue);
        this.writeEffects.set(func.$BoundTargetFunction, additionalFunctionEffects);
      } else {
        this.writeEffects.set(func, additionalFunctionEffects);
      }
    }
  }

  _optimizeReactComponentTreeBranches(reconciler, bindings, environmentRecordIdAfterGlobalCode) {
    let logger = this.moduleTracer.modules.logger;

    if (this.realm.react.verbose && reconciler.branchedComponentTrees.length > 0) {
      logger.logInformation(`  Evaluating React component tree branches...`);
    }
    // for now we just use abstract props/context, in the future we'll create a new branch with a new component
    // that used the props/context. It will extend the original component and only have a render method
    for (let _ref3 of reconciler.branchedComponentTrees) {
      let { rootValue: branchRootValue, evaluatedNode } = _ref3;

      let branchComponentType = (0, _utils.getComponentTypeFromRootValue)(this.realm, branchRootValue);
      if (branchComponentType === null) {
        evaluatedNode.status = "UNKNOWN_TYPE";
        continue;
      }
      // so we don't process the same component multiple times (we might change this logic later)
      if (reconciler.hasEvaluatedRootNode(branchComponentType, evaluatedNode)) {
        continue;
      }
      reconciler.clearComponentTreeState();
      let branchEffects = reconciler.renderReactComponentTree(branchComponentType, null, null, evaluatedNode);
      if (branchEffects === null) {
        if (this.realm.react.verbose) {
          logger.logInformation(`    ✖ ${evaluatedNode.name} (branch)`);
        }
        continue;
      }
      if (this._hasWriteConflictsFromReactRenders(bindings, branchEffects, [], evaluatedNode)) {
        if (this.realm.react.verbose) {
          logger.logInformation(`    ✖ ${evaluatedNode.name} (branch - write conflicts)`);
        }
        continue;
      }
      if (this.realm.react.verbose) {
        logger.logInformation(`    ✔ ${evaluatedNode.name} (branch)`);
      }
      let branchComponentTreeState = reconciler.componentTreeState;
      this._generateWriteEffectsForReactComponentTree(branchComponentType, branchEffects, branchComponentTreeState, evaluatedNode, environmentRecordIdAfterGlobalCode);
    }
  }

  _callOfFunction(funcValue) {
    let call = funcValue.$Call;
    (0, _invariant2.default)(call);
    let numArgs = funcValue.getLength();
    let args = [];
    (0, _invariant2.default)(funcValue instanceof _index.ECMAScriptSourceFunctionValue);
    let params = funcValue.$FormalParameters;
    if (numArgs && numArgs > 0 && params) {
      for (let parameterId of params) {
        if (t.isIdentifier(parameterId)) {
          // Create an AbstractValue similar to __abstract being called
          args.push(_index.AbstractValue.createAbstractArgument(this.realm, parameterId.name, funcValue.expressionLocation));
        } else {
          this.realm.handleError(new _errors.CompilerDiagnostic("Non-identifier args to additional functions unsupported", funcValue.expressionLocation, "PP1005", "FatalError"));
          throw new _errors.FatalError("Non-identifier args to additional functions unsupported");
        }
      }
    }

    let thisArg = _index.AbstractValue.createAbstractArgument(this.realm, "this", funcValue.expressionLocation, _index.ObjectValue);
    return call.bind(this, thisArg, args);
  }

  checkThatFunctionsAreIndependent(environmentRecordIdAfterGlobalCode) {
    let additionalFunctionsToProcess = this.__generateInitialAdditionalFunctions("__optimizedFunctions");
    // When we find declarations of nested optimized functions, we need to apply the parent
    // effects.
    let additionalFunctionStack = [];
    let additionalFunctions = new Set(additionalFunctionsToProcess.map(entry => entry.value));
    let optimizedFunctionsObject = this.moduleTracer.modules.logger.tryQuery(() => (0, _index2.Get)(this.realm, this.realm.$GlobalObject, "__optimizedFunctions"), this.realm.intrinsics.undefined);
    (0, _invariant2.default)(optimizedFunctionsObject instanceof _index.ObjectValue);

    // If there's an additional function that delcared functionValue, it must be
    // have already been evaluated for the __optimize call to have happened, so
    // this should always return either the defining additional function or void
    let getDeclaringAdditionalFunction = functionValue => {
      for (let [additionalFunctionValue, additionalEffects] of this.writeEffects) {
        // CreatedObjects is all objects created by this additional function but not
        // nested additional functions.
        let createdObjects = additionalEffects.effects.createdObjects;
        if (createdObjects.has(functionValue)) return additionalFunctionValue;
      }
    };

    let getEffectsFromAdditionalFunctionAndNestedFunctions = functionValue => {
      additionalFunctionStack.push(functionValue);
      (0, _invariant2.default)(functionValue instanceof _index.ECMAScriptSourceFunctionValue);
      let call = this._callOfFunction(functionValue);
      let effects = this.realm.evaluatePure(() => this.realm.evaluateForEffectsInGlobalEnv(call, undefined, "additional function"));
      (0, _invariant2.default)(effects);
      let additionalFunctionEffects = this._createAdditionalEffects(effects, true, "AdditionalFunctionEffects", environmentRecordIdAfterGlobalCode, getDeclaringAdditionalFunction(functionValue));
      (0, _invariant2.default)(additionalFunctionEffects);
      effects = additionalFunctionEffects.effects;
      this.writeEffects.set(functionValue, additionalFunctionEffects);

      // look for newly registered optimized functions
      let modifiedProperties = additionalFunctionEffects.effects.modifiedProperties;
      // Conceptually this will ensure that the nested additional function is defined
      // although for later cases, we'll apply the effects of the parents only.
      this.realm.withEffectsAppliedInGlobalEnv(() => {
        for (let [propertyBinding] of modifiedProperties) {
          let descriptor = propertyBinding.descriptor;
          if (descriptor && propertyBinding.object === optimizedFunctionsObject) {
            let newValue = descriptor.value;
            (0, _invariant2.default)(newValue instanceof _index.Value); //todo: this does not seem invariantly true
            let newEntry = this.__optimizedFunctionEntryOfValue(newValue);
            if (newEntry) {
              additionalFunctions.add(newEntry.value);
              getEffectsFromAdditionalFunctionAndNestedFunctions(newEntry.value);
              // Now we have to rember the stack of effects that need to be applied to deal with
              // this additional function.
            }
          }
        }
        return null;
      }, additionalFunctionEffects.effects);
      (0, _invariant2.default)(additionalFunctionStack.pop() === functionValue);
    };

    while (additionalFunctionsToProcess.length > 0) {
      let funcValue = additionalFunctionsToProcess.shift().value;
      getEffectsFromAdditionalFunctionAndNestedFunctions(funcValue);
    }
    (0, _invariant2.default)(additionalFunctionStack.length === 0);

    // check that functions are independent
    let conflicts = new Map();
    for (let fun1 of additionalFunctions) {
      (0, _invariant2.default)(fun1 instanceof _index.FunctionValue);
      let fun1Name = this.functionExpressions.get(fun1) || fun1.intrinsicName || "(unknown function)";
      // Also do argument validation here
      let additionalFunctionEffects = this.writeEffects.get(fun1);
      (0, _invariant2.default)(additionalFunctionEffects !== undefined);
      let e1 = additionalFunctionEffects.effects;
      (0, _invariant2.default)(e1 !== undefined);
      if (e1.result instanceof _completions.Completion && !e1.result instanceof _completions.PossiblyNormalCompletion) {
        let error = new _errors.CompilerDiagnostic(`Additional function ${fun1Name} may terminate abruptly`, e1.result.location, "PP1002", "FatalError");
        this.realm.handleError(error);
        throw new _errors.FatalError();
      }
      for (let fun2 of additionalFunctions) {
        if (fun1 === fun2) continue;
        (0, _invariant2.default)(fun2 instanceof _index.FunctionValue);
        let reportFn = () => {
          this.reportWriteConflicts(fun1Name, conflicts, e1.modifiedProperties, this._callOfFunction(fun2));
          return null;
        };
        let fun2Effects = this.writeEffects.get(fun2);
        (0, _invariant2.default)(fun2Effects);
        if (fun2Effects.parentAdditionalFunction) {
          let parentEffects = this.writeEffects.get(fun2Effects.parentAdditionalFunction);
          (0, _invariant2.default)(parentEffects);
          this.realm.withEffectsAppliedInGlobalEnv(reportFn, parentEffects.effects);
        } else {
          reportFn();
        }
      }
    }
    if (conflicts.size > 0) {
      for (let diagnostic of conflicts.values()) this.realm.handleError(diagnostic);
      throw new _errors.FatalError();
    }
  }

  getAdditionalFunctionValuesToEffects() {
    return this.writeEffects;
  }

  reportWriteConflicts(fname, conflicts, pbs, call2) {
    let reportConflict = location => {
      let error = new _errors.CompilerDiagnostic(`Property access conflicts with write in optimized function ${fname}`, location, "PP1003", "FatalError");
      conflicts.set(location, error);
    };
    let writtenObjects = new Set();
    pbs.forEach((val, key, m) => {
      writtenObjects.add(key.object);
    });
    let oldReportObjectGetOwnProperties = this.realm.reportObjectGetOwnProperties;
    this.realm.reportObjectGetOwnProperties = ob => {
      let location = this.realm.currentLocation;
      (0, _invariant2.default)(location);
      if (writtenObjects.has(ob) && !conflicts.has(location)) reportConflict(location);
    };
    let oldReportPropertyAccess = this.realm.reportPropertyAccess;
    this.realm.reportPropertyAccess = pb => {
      if (pb.object.refuseSerialization) return;
      let location = this.realm.currentLocation;
      if (!location) return; // happens only when accessing an additional function property
      if (pbs.has(pb) && !conflicts.has(location)) reportConflict(location);
    };
    try {
      (0, _errors2.ignoreErrorsIn)(this.realm, () => this.realm.evaluateForEffectsInGlobalEnv(call2));
    } finally {
      this.realm.reportPropertyAccess = oldReportPropertyAccess;
      this.realm.reportObjectGetOwnProperties = oldReportObjectGetOwnProperties;
    }
  }
}
exports.Functions = Functions; /**
                                * Copyright (c) 2017-present, Facebook, Inc.
                                * All rights reserved.
                                *
                                * This source code is licensed under the BSD-style license found in the
                                * LICENSE file in the root directory of this source tree. An additional grant
                                * of patent rights can be found in the PATENTS file in the same directory.
                                */
//# sourceMappingURL=functions.js.map